/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2018 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.cbs.perf.nas;

import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.Properties;
import java.util.UUID;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathException;
import javax.xml.xpath.XPathFactory;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsp.cbs.perf.config.NasAccess;
import dk.nsp.cbs.perf.http.HttpUtil;
import dk.nsp.nas.sms.seal.SealNamespaceContext;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.http.client.HttpClient;
import org.apache.log4j.Logger;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

// NOTE: This is not thread-safe!!!
public class DGWSService {

	private static Logger log = Logger.getLogger(DGWSService.class);

	private final String flowId = UUID.randomUUID().toString();

	private final String stsEndpoint;
	private final SOSIFactory sosiFactory;
	private final HttpClient httpClient;
	private final NasAccess nasAccess;

	private SystemIDCard stsSignedIdCard;

	public DGWSService(String stsEndpoint,
										 NasAccess nasAccess,
										 CredentialVault vault,
										 HttpClient httpClient,
										 Federation federation) {
		this.stsEndpoint = stsEndpoint;
		this.nasAccess = nasAccess;
		this.httpClient = httpClient;

		Properties properties = SignatureUtil.setupCryptoProviderForJVM();
		properties.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(true));
		sosiFactory = new SOSIFactory(federation, vault, properties);
	}

	public void ensureValidIdCard() {
		if (stsSignedIdCard != null) {
			Date expiryDate = stsSignedIdCard.getExpiryDate();
			if (expiryDate.toInstant().minus(5, ChronoUnit.HOURS).isAfter(Instant.now())) {
				// We have a valid idcard with at least 5 hours remaining.
				return;
			} else {
				log.info("Renewing IDCard from " + stsEndpoint);
			}
		} else {
			log.info("Getting first IDCard from " + stsEndpoint);
		}

		try {
			Document securityTokenRequest = buildSecurityTokenRequest();

			String requestXml = XmlUtil.node2String(securityTokenRequest, false, true);

			String responseXml = HttpUtil.issueRequest(httpClient, requestXml, "\"Issue\"", stsEndpoint).getBody();

			stsSignedIdCard = buildSystemIDCard(responseXml);

			log.info("Got new IDCard based on '" + stsSignedIdCard.getAlternativeIdentifier() + "' issued by " + stsSignedIdCard.getIssuer());
		} catch (Exception e) {
			throw new IllegalStateException(e);
		}
	}

	public void addHeaders(org.xmlsoap.schemas.soap.envelope.Header soapHeader, JAXBContext jaxbContext) throws XPathException, JAXBException {
		Request sosiRequest = sosiFactory.createNewRequest(false, flowId); // Flow id provided and Message ID generated
		sosiRequest.setIDCard(stsSignedIdCard);
		sosiRequest.setDGWSVersion("1.0.1");
		Document sosiRequestDocument = sosiRequest.serialize2DOMDocument();

		XPath xpath = XPathFactory.newInstance().newXPath();
		xpath.setNamespaceContext(new SealNamespaceContext());

		Node securityNode = (Node) xpath.compile("//soap:Header/wssec:Security").evaluate(sosiRequestDocument, XPathConstants.NODE);
		Security securityHeader = jaxbContext.createUnmarshaller().unmarshal(securityNode, Security.class).getValue();
		soapHeader.getAny().add(securityHeader);

		Node medcomHeaderNode = (Node) xpath.compile("//soap:Header/medcom:Header").evaluate(sosiRequestDocument, XPathConstants.NODE);
		Header medcomHeader = jaxbContext.createUnmarshaller().unmarshal(medcomHeaderNode, Header.class).getValue();
		soapHeader.getAny().add(medcomHeader);

	}

	private SystemIDCard buildSystemIDCard(String responseXml) {
		SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);
		SystemIDCard stsSignedIdCard = (SystemIDCard) securityTokenResponse.getIDCard();

		if (securityTokenResponse.isFault() || stsSignedIdCard == null) {
			// TODO: error handling
			log.error("We got a SOAP Fault from STS. The detailed message can be seen in the xml response");
			if (securityTokenResponse.isFault()) {
				log.error("A SOAP Fault was received: " + securityTokenResponse.getFaultString());
			}
			throw new IllegalStateException("SOAP Fault from STS");
		}
		return stsSignedIdCard;
	}

	private Document buildSecurityTokenRequest() {
		String idcard_username = null;
		String idcard_password = null;
		String idcard_alternative_identifier = null;

		CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, nasAccess.organizatonCode(),
																								 nasAccess.organizationName());
		X509Certificate certificate = sosiFactory.getCredentialVault().getSystemCredentialPair().getCertificate();
		SystemIDCard systemIDCard = sosiFactory.createNewSystemIDCard(nasAccess.systemName(), careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, idcard_username, idcard_password, certificate, idcard_alternative_identifier);

		SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
		securityTokenRequest.setIDCard(systemIDCard);
		Document doc = securityTokenRequest.serialize2DOMDocument();
		return doc;
	}
}
