/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2018 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsp.cbs.perf.nas;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigInteger;
import java.util.List;
import java.util.stream.Collectors;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.xpath.XPathException;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.advis.v10.NotifyContent;
import dk.nsp.cbs.perf.config.NasConfiguration;
import dk.nsp.cbs.perf.http.HttpUtil;
import dk.nsp.cbs.perf.http.Response;
import dk.nsp.cbs.perf.util.Xml;
import org.apache.http.client.HttpClient;
import org.apache.log4j.Logger;
import org.oasis_open.docs.wsn.b_2.CreatePullPoint;
import org.oasis_open.docs.wsn.b_2.CreatePullPointResponse;
import org.oasis_open.docs.wsn.b_2.DestroyPullPoint;
import org.oasis_open.docs.wsn.b_2.FilterType;
import org.oasis_open.docs.wsn.b_2.GetMessages;
import org.oasis_open.docs.wsn.b_2.GetMessagesResponse;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.Subscribe;
import org.oasis_open.docs.wsn.b_2.SubscribeResponse;
import org.oasis_open.docs.wsn.b_2.TopicExpressionType;
import org.oasis_open.docs.wsn.b_2.Unsubscribe;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3._2005._08.addressing.AttributedURIType;
import org.w3._2005._08.addressing.EndpointReferenceType;
import org.w3c.dom.Element;
import org.xmlsoap.schemas.soap.envelope.Body;
import org.xmlsoap.schemas.soap.envelope.Envelope;

public class NasService {

	private static Logger log = Logger.getLogger(NasService.class);

	private static final org.oasis_open.docs.wsn.b_2.ObjectFactory WSN_B2_FACTORY = new org.oasis_open.docs.wsn.b_2.ObjectFactory();

	private static final org.xmlsoap.schemas.soap.envelope.ObjectFactory SOAP_FACTORY = new org.xmlsoap.schemas.soap.envelope.ObjectFactory();

	private static final org.w3._2005._08.addressing.ObjectFactory WSA_FACTORY = new org.w3._2005._08.addressing.ObjectFactory();

	private final NasConfiguration configuration;
	private final DGWSService dgwsService; // NOTE: not thread safe
	private final HttpClient httpClient;

	private final JAXBContext jaxbContext;

	{
		try {
			this.jaxbContext = JAXBContext.newInstance(Envelope.class,
																								 Security.class,
																								 Header.class,
																								 CreatePullPoint.class,
																								 Subscribe.class,
																								 GetMessages.class,
																								 NotifyContent.class);
		} catch (JAXBException e) {
			log.error("Could not create jaxbContext!!!", e);
			throw new RuntimeException("Bad things are happening");
		}
	}

	public NasService(NasConfiguration configuration,
										DGWSService dgwsService,
										HttpClient httpClient) {
		this.configuration = configuration;
		this.dgwsService = dgwsService;
		this.httpClient = httpClient;
	}

	public Subscription subscribe(String topic) {
		try {
			dgwsService.ensureValidIdCard();

			CreatePullPoint createPullPointRequest = WSN_B2_FACTORY.createCreatePullPoint();
			CreatePullPointResponse createPullPointResponse = callService(SoapAction.CreatePullPoint,
																																		createPullPointRequest,
																																		configuration.pullpointFactoryEndpoint());
			String pullPointId = createPullPointResponse.getPullPoint().getAddress().getValue();

			log.info("Created new pullpoint: " + pullPointId);

			Subscribe subscribe = WSN_B2_FACTORY.createSubscribe();
			EndpointReferenceType endpointReferenceType = WSA_FACTORY.createEndpointReferenceType();
			AttributedURIType attributedURIType = WSA_FACTORY.createAttributedURIType();

			attributedURIType.setValue(pullPointId);
			endpointReferenceType.setAddress(attributedURIType);
			subscribe.setConsumerReference(endpointReferenceType);

			FilterType filterType = WSN_B2_FACTORY.createFilterType();
			TopicExpressionType topicExpressionType = WSN_B2_FACTORY.createTopicExpressionType();

			topicExpressionType.setDialect(configuration.topicExpressionDialect());
			topicExpressionType.getContent().add(topic);
			filterType.getAny().add(WSN_B2_FACTORY.createTopicExpression(topicExpressionType));
			subscribe.setFilter(filterType);

			SubscribeResponse subscribeResponse = callService(SoapAction.Subscribe,
																												subscribe,
																												configuration.subscriptionManagerEndpoint());
			String subscriptionId = subscribeResponse.getSubscriptionReference().getAddress().getValue();

			log.info("Created new subscription: " + subscriptionId);

			return new Subscription(-1, topic, pullPointId, subscriptionId);
		} catch (JAXBException | XPathException | IOException e) {
			throw new IllegalStateException(e);
		}
	}

	public void unsubscribe(String pullPointId, String subscriptionId) {
		try {
			dgwsService.ensureValidIdCard();

			Unsubscribe unsubscribe = WSN_B2_FACTORY.createUnsubscribe();
			callService(SoapAction.Unsubscribe, unsubscribe, subscriptionId);

			log.info("Removed subscription: " + subscriptionId);

			DestroyPullPoint destroyPullPoint = WSN_B2_FACTORY.createDestroyPullPoint();
			callService(SoapAction.DestroyPullPoint, destroyPullPoint, pullPointId);

			log.info("Removed pullpoint: " + pullPointId);
		} catch (JAXBException | XPathException | IOException e) {
			throw new IllegalStateException(e);
		}
	}

	public static class Job {
		public final String cpr;
		public final String jobId;
		private final boolean success;

		public Job(String cpr, String jobId, boolean success) {
			this.cpr = cpr;
			this.jobId = jobId;
			this.success = success;
		}

		public String cpr() {
			return cpr;
		}

		public String jobId() {
			return jobId;
		}

		public boolean isSuccess() {
			return success;
		}

		@Override
		public String toString() {
			return "Job{" +
				"cpr='" + cpr + '\'' +
				", jobId='" + jobId + '\'' +
				", success=" + success +
				'}';
		}
	}

	public List<Job> pullJobs(Subscription subscription) {
		try {
			dgwsService.ensureValidIdCard();

			GetMessages getMessages = WSN_B2_FACTORY.createGetMessages();
			getMessages.setMaximumNumber(BigInteger.valueOf(configuration.maxPullSize()));

			GetMessagesResponse getMessagesResponse = callService(SoapAction.GetMessages,
																														getMessages,
																														subscription.getPullpointId());

			List<Job> jobIds = getMessagesResponse.getNotificationMessage()
																														.stream()
																														.map(this::toJobId)
																														.collect(Collectors.toList());

			log.debug("Got " + jobIds.size() + " jobIds from " + subscription.getPullpointId());

			return jobIds;
		} catch (JAXBException | XPathException | IOException e) {
			throw new IllegalStateException(e);
		}
	}

	private static final String unknownJobId = "<unknown>";

	private Job toJobId(NotificationMessageHolderType notificationMessage) {
		final NotifyContent notifyContent = (NotifyContent) notificationMessage.getMessage().getAny();
		final String jobId;
		final boolean success;
		if (notifyContent.getAny() instanceof Element) {
			jobId = Xml.childText((Element) notifyContent.getAny(), "JobId")
								 .orElse(unknownJobId);
			success = Xml.childText((Element) notifyContent.getAny(), "JobStatus")
									 .map("OK"::equals)
									 .orElse(false);
		} else {
			jobId = unknownJobId;
			success = false;
		}
		return new Job(notifyContent.getId(), jobId, success);
	}

	@SuppressWarnings("unchecked")
	private <RequestType, ResponseType> ResponseType callService(SoapAction soapAction,
																															 RequestType request,
																															 String endpoint) throws JAXBException, XPathException, IOException {
		Body soapBody = SOAP_FACTORY.createBody();
		soapBody.getAny().add(request);
		org.xmlsoap.schemas.soap.envelope.Header soapHeader = SOAP_FACTORY.createHeader();
		Envelope soapEnvelope = SOAP_FACTORY.createEnvelope();
		dgwsService.addHeaders(soapHeader, jaxbContext);
		soapEnvelope.setHeader(soapHeader);
		soapEnvelope.setBody(soapBody);

		StringWriter requestXMLWriter = new StringWriter();
		jaxbContext.createMarshaller().marshal(soapEnvelope, requestXMLWriter);
		String requestXml = requestXMLWriter.toString();

		final Response response = HttpUtil.issueRequest(httpClient, requestXml, soapAction.value(), endpoint);
		if (response.getCode() != 200) {
			throw new RuntimeException("Call to "+endpoint+" failed; "+response.getBody());
		}
		final String responseXml = response.getBody();
		final Envelope responseEnvelope = (Envelope) jaxbContext.createUnmarshaller().unmarshal(new StringReader(responseXml));

		try {
			return (ResponseType) responseEnvelope.getBody().getAny().get(0);
		} catch (Exception e) {
			throw new RuntimeException("Unable to cast body.", e);
		}
	}
}
