package dk.nsp.cbs.perf.runner;

import dk.nsp.cbs.perf.config.TestDriverConfig;
import dk.nsp.cbs.perf.nas.NasService;
import dk.nsp.cbs.perf.nas.Subscription;
import dk.nsp.cbs.perf.service.CbsClient;
import dk.nsp.cbs.perf.service.StsJwtClient;
import dk.sosi.seal.model.IdentityToken;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedTransferQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * Optional system properties:
 * <ul>
 *   <li>test.nas.poll.initial.delay (default 0) - denotes the initial delay in milliseconds before starting a poll</li>
 *   <li>test.nas.poll.period (default 500) - denotes the interval in milliseconds in which to poll the NAS for JobIds</li>
 *   <li>test.feed.initial.delay (default 0) - denotes the initial delay in milliseconds before starting to check the queue</li>
 *   <li>test.feed.period (default 500) - denotes the interval in milliseconds in which to check and perhaps feed the queue with data</li>
 *   <li>test.transit.queue.max.drain (default 1000)- denotes the maximum number of entities to drain from the transit queue at each attempt</li>
 *   <li>test.stats.initial.delay (default 60) - denotes the initial delay in seconds before starting to show stats</li>
 *   <li>test.stats.period (default 600) - denotes the number of seconds between each output of stats</li>
 *   <li>test.skip.cbs (default false) - enable or disable calls to CBS from threads in thread pool</li>
 * </ul>
 *
 * @author jns, Arosii Information Systems A/S
 */
public class Conductor {

  private final CbsClient cbsClient;
  private final StsJwtClient stsJwtClient;
  private final NasService nasService;
  private final Subscription subscription;
  private final TestDriverConfig conf;
  private final TestDriver.InData in = new TestDriver.InData();

  private final ScheduledExecutorService feed = Executors.newScheduledThreadPool(1);
  private final ScheduledExecutorService cbs;
  private final ScheduledExecutorService nas = Executors.newScheduledThreadPool(1);
  private final ScheduledExecutorService stat = Executors.newScheduledThreadPool(1);

  private final LinkedTransferQueue<String> feederQueue = new LinkedTransferQueue<>();
  private final LinkedTransferQueue<TestDriver.TransitElement> transitQueue = new LinkedTransferQueue<>();

  private final Logger logger = LoggerFactory.getLogger(getClass());

  private final static AtomicLong processedForCaching = new AtomicLong(0);
  private final static AtomicLong servedFromCache = new AtomicLong(0);
  private final static AtomicLong totalProcessed = new AtomicLong(0);

  public Conductor(CbsClient cbsClient, StsJwtClient stsJwtClient, NasService nasService,
                   Subscription subscription, TestDriverConfig testDriverConfig) {
    this.cbsClient = cbsClient;
    this.stsJwtClient = stsJwtClient;
    this.nasService = nasService;
    this.subscription = subscription;
    this.conf = testDriverConfig;
    this.cbs = Executors.newScheduledThreadPool(testDriverConfig.numThreads());

    Runtime.getRuntime().addShutdownHook(new Thread(() -> {
      try {
        feed.shutdown();
        feed.awaitTermination(3, TimeUnit.SECONDS);
        cbs.shutdown();
        cbs.awaitTermination(3, TimeUnit.SECONDS);
        nas.shutdown();
        nas.awaitTermination(3, TimeUnit.SECONDS);
        stat.shutdown();
        stat.awaitTermination(1, TimeUnit.SECONDS);
      } catch (InterruptedException e) {
      }
    }));
  }

  public void play() throws IOException {
    IntStream
        .rangeClosed(1, conf.numThreads())
        .forEach(i ->
            cbs.scheduleAtFixedRate(new Taker(feederQueue, transitQueue),
                new Random().nextInt((int) (conf.requestWindowSize() + 1)),
                conf.requestWindowSize(),
                TimeUnit.MILLISECONDS)
        );

    nas.scheduleAtFixedRate(new Poller(transitQueue, nasService, subscription),
        Long.getLong("test.nas.poll.initial.delay", 0),
        Long.getLong("test.nas.poll.period", 500),
        TimeUnit.MILLISECONDS);
    feed.scheduleAtFixedRate(new Feeder(feederQueue),
        Long.getLong("test.feed.initial.delay",0),
        Long.getLong("test.feed.period", 500),
        TimeUnit.MILLISECONDS);
    stat.scheduleAtFixedRate(new Stats(),
        Long.getLong("test.stats.initial.delay", 60),
        Long.getLong("test.stats.period", 600),
        TimeUnit.SECONDS);
    logger.info("Started " + conf.numThreads() + " CBS call threads.");
  }

  /* calls cbs */
  private class Taker implements Runnable {
    private final LinkedTransferQueue<String> feedQueue;
    private final LinkedTransferQueue<TestDriver.TransitElement> transitQueue;

    public Taker(LinkedTransferQueue<String> feedQueue, LinkedTransferQueue<TestDriver.TransitElement> transitQueue) {
      this.feedQueue = feedQueue;
      this.transitQueue = transitQueue;
    }

    private final String certificateInfo = "SubjectDN={SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK},IssuerDN={CN=TRUST2408 Systemtest XIX CA, O=TRUST2408, C=DK},CertSerial={1393564623}";

    @Override
    public void run() {
      final String pnr = feedQueue.poll();

      if (pnr == null) {
        return;
      }

      final IdentityToken identityToken;
      try {
        identityToken = stsJwtClient.call(pnr, certificateInfo);
      } catch (Exception e) {
        logger.error("Error calling STS with " + pnr + ":", e);
        feederQueue.add(pnr);
        return;
      }

      if (System.getProperty("test.skip.cbs") != null) {
        return;
      }

      final CbsClient.CbsResponse res;
      try {
        res = cbsClient.call(identityToken, pnr);
      } catch (Exception e) {
        logger.error("Error calling CBS with " + pnr + ":", e);
        feederQueue.add(pnr);
        return;
      }

      totalProcessed.getAndIncrement();
      if (res.isHit()) {
        servedFromCache.getAndIncrement();
      }
      if (res.getJobId() != null) {
        transitQueue.add(new TestDriver.TransitElement(res.getJobId(), pnr));
        logger.debug("MISS:" + res.getJobId() + ", " + pnr);
        processedForCaching.getAndIncrement();
      }
    }
  }

  /* calls NAS */
  private class Poller implements Runnable {
    private final LinkedTransferQueue<TestDriver.TransitElement> transitQueue;
    private final NasService nasService;
    private final Subscription subscription;

    public Poller(LinkedTransferQueue<TestDriver.TransitElement> transitQueue, NasService nasService, Subscription subscription) {
      this.transitQueue = transitQueue;
      this.nasService = nasService;
      this.subscription = subscription;
    }

    @Override
    public void run() {
      final List<NasService.Job> jobs = nasService.pullJobs(subscription)
                                            .stream()
                                            .peek(job -> logger.debug("NAS:" + job.jobId() + ", " + job.cpr()))
                                            .collect(Collectors.toList());
      final List<TestDriver.TransitElement> inTransit = new ArrayList<>();
      transitQueue.drainTo(inTransit, Integer.getInteger("test.transit.queue.max.drain", 1000));

      final List<TestDriver.TransitElement> inCache = inTransit.stream()
          .filter(transitElement -> jobs.stream()
              .anyMatch(job -> transitElement.getJobId().equals(job.jobId())))
          .collect(Collectors.toList());
      inTransit.removeAll(inCache);
      transitQueue.addAll(inTransit); // feed back all not yet in cache
    }
  }

  /* feeds the PNR queue when empty */
  private class Feeder implements Runnable {
    private final LinkedTransferQueue<String> feederQueue;
    private final List<String> pnrs = new ArrayList<>();

    public Feeder(LinkedTransferQueue<String> feederQueue) throws IOException {
      this.feederQueue = feederQueue;
      Files.lines(in.path())
          .forEach(pnrs::add);
      Collections.shuffle(pnrs);
    }

    @Override
    public void run() {
      if (feederQueue.isEmpty()) {
        feederQueue.addAll(pnrs);
        logger.info("Feed queue had " + pnrs.size() + " elements injected.");
      }
    }
  }

  /* output statistics */
  private class Stats implements Runnable {
    @Override
    public void run() {
      final long served = servedFromCache.get();
      final long processed = processedForCaching.get();
      final long total = totalProcessed.get();
      final double hitPrc;
      final double missPrc;
      if (total == 0) {
        hitPrc = Double.NaN;
        missPrc = Double.NaN;
      } else {
        hitPrc = Math.round(((((double)served / (double)total) * 100d) * 100d) / 100d);
        missPrc = Math.round(((((double)processed / (double)total) * 100d) * 100d) / 100d);
      }
      logger.info("Statistics [h:" + hitPrc + " m:" + missPrc + ",total-processed:" + total + ",served-from-cache:" + served + ",processed-for-caching:" + processed + ",feeder-queue:" + feederQueue.size() + ",transit-queue:" + transitQueue.size() + "]");
    }
  }
}