package dk.nsp.cbs.perf.runner;

import dk.nsp.cbs.perf.config.TestDriverConfig;
import dk.nsp.cbs.perf.nas.NasService;
import dk.nsp.cbs.perf.nas.Subscription;
import dk.nsp.cbs.perf.service.CbsClient;
import dk.nsp.cbs.perf.service.StsJwtClient;
import dk.sosi.seal.model.IdentityToken;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedTransferQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.IntStream;

/**
 * Optional system properties:
 * <ul>
 *   <li>rampup.nas.drain.retry (default 500) - denotes the pause between calling NAS during ramp up</li>
 *   <li>rampup.stats.initial.delay (default 600) - denotes the initial number of seconds before any statistics output</li>
 *   <li>rampup.stats.period (default 600) - denotes the number of seconds between statistics output</li>
 * </ul>
 *
 * @author jns, Arosii Information Systems A/S
 */
public class RampUp {
  private final CbsClient cbsClient;
  private final StsJwtClient stsJwtClient;
  private final NasService nasService;
  private final Subscription subscription;
  private final TestDriver.InData in = new TestDriver.InData();
  private final TestDriver.Distribution dist;
  private final ExecutorService cbs = Executors.newFixedThreadPool(Integer.getInteger("rampup.cbs.threads", 20));
  private final ExecutorService nas = Executors.newFixedThreadPool(Integer.getInteger("rampup.nas.threads", 1));
  private final ScheduledExecutorService stat = Executors.newScheduledThreadPool(1);

  private final Logger logger = LoggerFactory.getLogger(getClass());
  private final AtomicLong total = new AtomicLong();
  private final AtomicLong cached = new AtomicLong();
  private final AtomicLong uncached = new AtomicLong();

  public RampUp(CbsClient cbsClient, StsJwtClient stsJwtClient,
                NasService nasService, Subscription subscription, TestDriverConfig testDriverConfig) {
    this.cbsClient = cbsClient;
    this.stsJwtClient = stsJwtClient;
    this.nasService = nasService;
    this.subscription = subscription;
    this.dist = new TestDriver.Distribution(testDriverConfig.dist());

    Runtime.getRuntime().addShutdownHook(new Thread(() -> {
      try {
        cbs.shutdown();
        cbs.awaitTermination(3, TimeUnit.SECONDS);
        nas.shutdown();
        nas.awaitTermination(3, TimeUnit.SECONDS);
        stat.shutdown();
        stat.awaitTermination(1, TimeUnit.SECONDS);
      } catch (InterruptedException e) {
      }
    }));

  }

  private final String certificateInfo = "SubjectDN={SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK},IssuerDN={CN=TRUST2408 Systemtest XIX CA, O=TRUST2408, C=DK},CertSerial={1393564623}";

  public void execute() throws IOException, InterruptedException {
    final LinkedTransferQueue<String> pnrs = new LinkedTransferQueue<>();
    final LinkedTransferQueue<String> transit = new LinkedTransferQueue<>();

    Files.lines(in.path())
        .forEach(pnrs::add);
    logger.info("Ramp up started");

    stat.scheduleAtFixedRate(() -> {
          logger.info("Statistics [t:" + total.longValue() + ",uc:" + uncached.longValue() + ",c:" + cached.longValue() + "]");
        },
        Long.getLong("rampup.stats.initial.delay", 600),
        Long.getLong("rampup.stats.period", 600),
        TimeUnit.SECONDS);

    IntStream
        .rangeClosed(1, pnrs.size())
        .forEach(i -> {
          cbs.submit(() -> {
            String pnr = pnrs.poll();
            total.getAndIncrement();
            if (dist.putInCache()) {
              try {
                final IdentityToken identityToken = stsJwtClient.call(pnr, certificateInfo);
                final CbsClient.CbsResponse response = cbsClient.call(identityToken, pnr);
                if (response.isHit()) {
                  logger.warn(pnr + " was hit.");
                  return;
                }
                transit.add(response.getJobId());
              } catch (Exception e) {
                logger.error("Error calling CBS or STS with " + pnr + ":", e);
                pnrs.add(pnr);
                return;
              }
              cached.getAndIncrement();
            } else {
              uncached.getAndIncrement();
            }
          });
        });

    // wait until backend is drained...
    nas.submit(() -> {
      while (!transit.isEmpty()) {
        nasService
            .pullJobs(subscription)
            .stream()
            .map(NasService.Job::jobId)
            .forEach(transit::remove);
          try {
            Thread.sleep(Long.getLong("rampup.nas.drain.retry", 500));
          } catch (InterruptedException e) {
          }
      }
    });

    cbs.shutdown();
    cbs.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS); // let threads finish...

    nas.shutdown();
    nas.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);

    stat.shutdown();
    stat.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);

    logger.info("Ramp up ended [cached " + cached + ", uncached " + uncached + "]");
  }
}
