package dk.nsp.cbs.perf.runner;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;

import dk.nsp.cbs.perf.config.CredentialVaultInfo;
import dk.nsp.cbs.perf.config.NasAccess;
import dk.nsp.cbs.perf.config.NasConfiguration;
import dk.nsp.cbs.perf.nas.DGWSService;
import dk.nsp.cbs.perf.nas.NasService;
import dk.nsp.cbs.perf.nas.Subscription;
import dk.nsp.cbs.perf.service.CbsClient;
import dk.nsp.cbs.perf.service.StsJwtClient;
import dk.nsp.cbs.perf.service.vault.VaultService;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVault;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.DefaultHttpClient;
import org.cfg4j.provider.ConfigurationProvider;
import org.cfg4j.provider.ConfigurationProviderBuilder;
import org.cfg4j.source.context.environment.ImmutableEnvironment;
import org.cfg4j.source.files.FilesConfigurationSource;

public class Runner {
  public static void main(String[] args) {
    // Configuration; http://www.cfg4j.org/releases/latest/#users-guide

    // etc/local is running against a locally started stub server
    final String envName = System.getProperty("perf.env", "etc/dev"); // etc/dev is also possible for testing against the external test setup
    final ConfigurationProvider cfgProvider = new ConfigurationProviderBuilder()
      .withConfigurationSource(new FilesConfigurationSource()) // we can also merge configuration files
      .withEnvironment(new ImmutableEnvironment(envName))
      .build();

    final String vaultDirStr = cfgProvider.getProperty("vaultDir", String.class);
    final CredentialVaultInfo idpVaultInfo = cfgProvider.bind("idpVault", CredentialVaultInfo.class);
    final CredentialVaultInfo signingVaultInfo = cfgProvider.bind("signingVault", CredentialVaultInfo.class);

    final String stsJwtUrl = cfgProvider.getProperty("sts.jwt.location", String.class);
    final String stsDgwsUrl = cfgProvider.getProperty("sts.dgws.location", String.class);
    final String cbsUrl = cfgProvider.getProperty("cbs.location", String.class);

    final String stsJwtAuduence = cfgProvider.getProperty("sts.jwt.audience", String.class);
    final String stsJwtKid = cfgProvider.getProperty("sts.jwt.kid", String.class);
    final String stsJwtScope = cfgProvider.getProperty("sts.jwt.scope", String.class);

    final NasConfiguration nasConfiguration = cfgProvider.bind("nas", NasConfiguration.class);
    final NasAccess nasAccess = cfgProvider.bind("nasAccess", NasAccess.class);
    final CredentialVaultInfo nasAccessVaultInfo = cfgProvider.bind("nasAccess.vault", CredentialVaultInfo.class);

    final String nasTopic = cfgProvider.getProperty("nas.topic", String.class);

    // For testing configuration setup
//    if (true)
//      return;

    // Setup

    final Path vaultDir = Paths.get(vaultDirStr);
    final HttpClient httpClient = new DefaultHttpClient();
    final VaultService vaultService = new VaultService(vaultDir::resolve);

    final CredentialVault idpVault = vaultService.getVaultFromResource(idpVaultInfo);
    final CredentialVault signingVault = vaultService.getVaultFromResource(signingVaultInfo);
    final CredentialVault nasVault = vaultService.getVaultFromResource(nasAccessVaultInfo);

    final Federation federation = new SOSITestFederation(new Properties());

    final StsJwtClient stsJwtClient = new StsJwtClient(stsJwtUrl,
                                                       httpClient,
                                                       stsJwtAuduence,
                                                       stsJwtKid,
                                                       idpVault,
                                                       signingVault,
                                                       stsJwtScope);

    final CbsClient cbsClient = CbsClient.create(cbsUrl, httpClient, federation, signingVault);

    final NasService nasService = new NasService(nasConfiguration,
                                                 new DGWSService(stsDgwsUrl,
                                                                 nasAccess,
                                                                 nasVault,
                                                                 httpClient,
                                                                 federation),
                                                 httpClient);

    // final Subscription subscription = nasService.subscribe(nasTopic);
    // final List<String> strings = nasService.pullJobIds(subscription);

    final Subscription subscription = nasService.subscribe(nasConfiguration.topic());
    System.out.println("Subscription is" + subscription.getPullpointId() + "," + subscription.getSubscriptionId());

    new TestRunner(stsJwtClient, cbsClient, "0501792279").run();

    try {
      Thread.sleep(30000);
    } catch (InterruptedException e) {
      e.printStackTrace();
    }

    try {
      System.out.println("From NAS:" + nasService.pullJobs(subscription));
    } catch (Exception e) {
      System.err.println("NAS pulling failed.");
      e.printStackTrace();
    }

    try {
      System.in.read();
    } catch (IOException e) {
      e.printStackTrace();
    }

//    nasService.unsubscribe("http://test1.ekstern-test.nspop.dk:8080/nas2/pullpoint/service/0c78e1c6-d53c-420c-9bc3-c2c1c68dbd2b",
//                           "http://test1.ekstern-test.nspop.dk:8080/nas2/subscriptionmanager/service/0e1421f2-0e3d-445e-9235-2534d98855e4");

    nasService.unsubscribe(subscription.getPullpointId(), subscription.getSubscriptionId());
  }

  public static class TestRunner {

    private final StsJwtClient stsJwtClient;
    private final CbsClient cbsClient;
    private final String cpr;

    public TestRunner(StsJwtClient stsJwtClient,
                      CbsClient cbsClient,
                      String cpr) {
      this.stsJwtClient = stsJwtClient;
      this.cbsClient = cbsClient;
      this.cpr = cpr;
    }

    public void run() {
      final IdentityToken identityToken
        = stsJwtClient.call(cpr,
                            "SubjectDN={SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK},IssuerDN={CN=TRUST2408 Systemtest XIX CA, O=TRUST2408, C=DK},CertSerial={1393564623}");

      final CbsClient.CbsResponse response = cbsClient.call(identityToken, cpr);
      System.out.println("Got a hit: " + response.isHit() + "[jobId: " + response.getJobId() + "]");
    }
  }
}
