package dk.nsp.cbs.perf.runner;

import java.io.IOException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.ThreadLocalRandom;

import dk.nsp.cbs.perf.config.CredentialVaultInfo;
import dk.nsp.cbs.perf.config.NasAccess;
import dk.nsp.cbs.perf.config.NasConfiguration;
import dk.nsp.cbs.perf.config.TestDriverConfig;
import dk.nsp.cbs.perf.nas.DGWSService;
import dk.nsp.cbs.perf.nas.NasService;
import dk.nsp.cbs.perf.nas.Subscription;
import dk.nsp.cbs.perf.service.CbsClient;
import dk.nsp.cbs.perf.service.StsJwtClient;
import dk.nsp.cbs.perf.service.vault.VaultService;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVault;
import org.apache.http.HttpHost;
import org.apache.http.client.HttpClient;
import org.apache.http.conn.routing.HttpRoute;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.PoolingClientConnectionManager;
import org.cfg4j.provider.ConfigurationProvider;
import org.cfg4j.provider.ConfigurationProviderBuilder;
import org.cfg4j.source.context.environment.ImmutableEnvironment;
import org.cfg4j.source.files.FilesConfigurationSource;

/**
 * @author jns, Arosii Information Systems A/S
 *
 * Get some personnumre from SDS database:
 * select distinct CPR from Person order by rand() limit 1000 into outfile 'rand-cpr.txt' lines terminated by '\n';
 *
 * Optional system properties:
 * <ul>
 *   <li>test.http.connman.max.total (default 200) - denotes the maximum number of http connections across all endpoints</li>
 *   <li>test.http.connman.stsjwt.max (default 95) - denotes the maximum number of http connections to the STS JWT2IDWS endpoint</li>
 *   <li>test.http.connman.cbs.max (default 95) - denotes the maximum number of http connections to the CBS endpoint</li>
 *   <li>test.http.connman.stsdgws.max (default 10) - denotes the maximum number of http connections to the STS DGWS signing endpoint</li>
 *   <li>test.sts.timing (default false) - enable or disable timing of calls to STS</li>
 * </ul>
 */
public class Smoke {

  public static void main(String[] args) throws IOException, InterruptedException {
    // Configuration; http://www.cfg4j.org/releases/latest/#users-guide
//    final String envName = System.getProperty("perf.env", ".");
    final String envName = System.getProperty("perf.env", "etc/dev"); // etc/dev is also possible for testing against the external test setup
    final ConfigurationProvider cfgProvider = new ConfigurationProviderBuilder()
        .withConfigurationSource(new FilesConfigurationSource()) // we can also merge configuration files
        .withEnvironment(new ImmutableEnvironment(envName))
        .build();

    final String vaultDirStr = cfgProvider.getProperty("vaultDir", String.class);
    final CredentialVaultInfo idpVaultInfo = cfgProvider.bind("idpVault", CredentialVaultInfo.class);
    final CredentialVaultInfo signingVaultInfo = cfgProvider.bind("signingVault", CredentialVaultInfo.class);

    final String stsJwtUrl = cfgProvider.getProperty("sts.jwt.location", String.class);
    final String stsDgwsUrl = cfgProvider.getProperty("sts.dgws.location", String.class);
    final String cbsUrl = cfgProvider.getProperty("cbs.location", String.class);

    final String stsJwtAuduence = cfgProvider.getProperty("sts.jwt.audience", String.class);
    final String stsJwtKid = cfgProvider.getProperty("sts.jwt.kid", String.class);
    final String stsJwtScope = cfgProvider.getProperty("sts.jwt.scope", String.class);

    final NasConfiguration nasConfiguration = cfgProvider.bind("nas", NasConfiguration.class);
    final NasAccess nasAccess = cfgProvider.bind("nasAccess", NasAccess.class);
    final CredentialVaultInfo nasAccessVaultInfo = cfgProvider.bind("nasAccess.vault", CredentialVaultInfo.class);

    final TestDriverConfig testDriverConfig = cfgProvider.bind("drv", TestDriverConfig.class);

    final Path vaultDir = Paths.get(vaultDirStr);

    final PoolingClientConnectionManager connectionManager = new PoolingClientConnectionManager();
    connectionManager.setMaxTotal(Integer.getInteger("test.http.connman.max.total", 200));
    connectionManager.setMaxPerRoute(new HttpRoute(new HttpHost(new URL(stsJwtUrl).getHost())),Integer.getInteger("test.http.connman.stsjwt.max", 95));
    connectionManager.setMaxPerRoute(new HttpRoute(new HttpHost(new URL(cbsUrl).getHost())),Integer.getInteger("test.http.connman.cbs.max", 95));
    connectionManager.setMaxPerRoute(new HttpRoute(new HttpHost(new URL(stsDgwsUrl).getHost())),Integer.getInteger("test.http.connman.stsdgws.max", 10));
    final HttpClient httpClient = new DefaultHttpClient(connectionManager);

    final VaultService vaultService = new VaultService(vaultDir::resolve);

    final CredentialVault idpVault = vaultService.getVaultFromResource(idpVaultInfo);
    final CredentialVault signingVault = vaultService.getVaultFromResource(signingVaultInfo);
    final CredentialVault nasVault = vaultService.getVaultFromResource(nasAccessVaultInfo);

    final Federation federation = new SOSITestFederation(new Properties());

    final StsJwtClient stsJwtClient = new StsJwtClient(stsJwtUrl,
        httpClient,
        stsJwtAuduence,
        stsJwtKid,
        idpVault,
        signingVault,
        stsJwtScope);

    final CbsClient cbsClient = new CbsClient(cbsUrl, httpClient, federation, signingVault, System.out::println);

    final NasService nasService = new NasService(nasConfiguration,
        new DGWSService(stsDgwsUrl,
            nasAccess,
            nasVault,
            httpClient,
            federation),
        httpClient);

    final Subscription subscription = nasService.subscribe(nasConfiguration.topic());
    try {
      final String cpr = System.getProperty("smoke.cpr", "1502900230");
      System.out.println("Calling with CPR: "+cpr);
      final CbsClient.CbsResponse response = new TestRunner(stsJwtClient, cbsClient, cpr).run();
      if (response == null) {
        return;
      }
      System.out.println("Got a hit: " + response.isHit() + "[jobId: " + response.getJobId() + "]");
      if (response.isHit()) {
        return;
      }

      while (true) {
        try {
          System.out.println("Waiting 3 second for the job to complete.");
          Thread.sleep(3000);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }

        final List<NasService.Job> jobs;
        try {
          jobs = nasService.pullJobs(subscription);
          System.out.println("From NAS:" + jobs);
        } catch (Exception e) {
          System.err.println("NAS pulling failed.");
          e.printStackTrace();
          return;
        }

        if (jobs.size() > 0 && !jobs.get(0).isSuccess()) {
          System.out.println("Failure from NAS on Job: "+jobs.get(0).jobId());
          return;
        }

        if (jobs.size() > 0) {
          System.out.println("Found from NAS: " + jobs);
          final CbsClient.CbsResponse responseAfterNas = new TestRunner(stsJwtClient, cbsClient, cpr).run();
          if (responseAfterNas == null) {
            return;
          }
          System.out.println("Got a hit: " + responseAfterNas.isHit() + "[jobId: " + responseAfterNas.getJobId() + "]");
          break;
        }
      }
    } finally {
      nasService.unsubscribe(subscription.getPullpointId(), subscription.getSubscriptionId());
    }
  }

  public static class TestRunner {

    private final StsJwtClient stsJwtClient;
    private final CbsClient cbsClient;
    private final String cpr;

    public TestRunner(StsJwtClient stsJwtClient,
                      CbsClient cbsClient,
                      String cpr) {
      this.stsJwtClient = stsJwtClient;
      this.cbsClient = cbsClient;
      this.cpr = cpr;
    }

    public CbsClient.CbsResponse run() {
      final String certificateInfo = "SubjectDN={SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK},IssuerDN={CN=TRUST2408 Systemtest XIX CA, O=TRUST2408, C=DK},CertSerial={1393564623}";
      try {
        final IdentityToken identityToken = stsJwtClient.call(cpr, certificateInfo);
        return cbsClient.call(identityToken, cpr);
      } catch (Exception e) {
        System.err.println("Error calling STS or CBS: ");
        e.printStackTrace();
        return null;
      }
    }
  }
}