package dk.nsp.cbs.perf.service;

import java.util.Optional;
import java.util.Properties;
import java.util.UUID;
import java.util.function.Consumer;

import dk.nsp.cbs.perf.util.Xml;
import dk.nsp.cbs.perf.http.HttpUtil;
import dk.nsp.cbs.perf.http.Response;
import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.LibertyRequestDOMEnhancer;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.http.client.HttpClient;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class CbsClient {

  private final String endpoint;
  private final HttpClient httpClient;
  private final OIOIDWSFactory oioidwsFactory;
  private final Consumer<String> responseConsumer;

  public CbsClient(String endpoint, HttpClient httpClient, Federation federation, CredentialVault signingVault, Consumer<String> responseConsumer) {
    this.endpoint = endpoint;
    this.httpClient = httpClient;
    this.oioidwsFactory = new OIOIDWSFactory(federation, signingVault);
    this.responseConsumer = responseConsumer;
  }

  public static CbsClient create(String endpoint, HttpClient httpClient, Federation federation, CredentialVault signingVault) {
    return new CbsClient(endpoint, httpClient, federation, signingVault, null);
  }

  private final String soapBody = "<soapenv:Envelope xmlns:soapenv=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:cor=\"http://sundhedsdatastyrelsen.dk/coronapass/2021/03/coronapass-web-service\">\n" +
    "   <soapenv:Header/>\n" +
    "   <soapenv:Body>\n" +
    "      <cor:GetCoronaPassRequest>\n" +
    "         <cor:Identifier>%s</cor:Identifier>\n" +
    "      </cor:GetCoronaPassRequest>\n" +
    "   </soapenv:Body>\n" +
    "</soapenv:Envelope>";

  public interface CbsResponse {
    boolean isHit();
    default String getJobId() {
      return null;
    }
  }

  public static CbsResponse parseCbsResponse(String str) {

    final Document document = XmlUtil.readXml(new Properties(),
                                              str,
                                              false);
    final Element root = document.getDocumentElement();
    final Optional<Element> response = Xml.children(root, "Body")
                                          .findFirst()
                                          .flatMap(body -> Xml.children(body).findFirst())
                                          .flatMap(resp -> Xml.children(resp).findFirst());

    switch (response.get().getTagName()) {
      case "CoronaPasses":
        return () -> true;
      case "FutureResult":
        return new CbsResponse() {
          @Override
          public boolean isHit() {
            return false;
          }

          @Override
          public String getJobId() {
            return Xml.childText(response.get(), "JobId").get();
          }
        };
      default:
        return null;
    }
  }
  public CbsResponse call(IdentityToken identityToken, String cpr) {
    final Document soapBodyDoc = XmlUtil.readXml(new Properties(),
                                                 String.format(soapBody, cpr),
                                                 false);

    signDocument(soapBodyDoc, identityToken);

    final String bodyToSend = XmlUtil.node2String(soapBodyDoc);
    final Response response = HttpUtil.issueRequest(httpClient,
                                                    bodyToSend,
                                                    "",
                                                    endpoint);
    if (response.getCode() != 200) {
      throw new RuntimeException("Calling CBS failed: "+response.getCode()+"\n"+response.getBody());
    }

    if (responseConsumer != null) {
      responseConsumer.accept(response.getBody());
    }

    return parseCbsResponse(response.getBody());
  }

  private void signDocument(Document document, IdentityToken identityToken) {
    final LibertyRequestDOMEnhancer enhancer = oioidwsFactory.createRequestDOMEnhancer(document);
    enhancer.setIdentityToken(identityToken);
    enhancer.setWSAddressingMessageID(UUID.randomUUID().toString());
    enhancer.setWSAddressingAction("http://nspop.dk/2020/03/19/accesshandler/test#RequestServiceCall");
    enhancer.enhanceAndSign();
  }
}
