package dk.nsp.cbs.perf.service;

import java.time.Duration;
import java.time.Instant;
import java.util.Date;
import java.util.Properties;

import dk.nsp.cbs.perf.http.HttpUtil;
import dk.nsp.cbs.perf.http.Response;
import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.model.JWTToIdentityTokenResponse;
import dk.sosi.seal.model.dombuilders.JWTToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.CertificateInfo;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import io.jsonwebtoken.Claims;
import io.jsonwebtoken.JwtBuilder;
import io.jsonwebtoken.Jwts;
import io.jsonwebtoken.SignatureAlgorithm;
import org.apache.http.client.HttpClient;
import org.w3c.dom.Document;

public class StsJwtClient {

  private final String endpoint;
  private final HttpClient httpClient;
  private final String jwt2IdwsAudience;
  private final String kid;
  private final CredentialVault idp;
  private final CredentialVault signingVault;
  private final String scope;

  public StsJwtClient(String endpoint,
                      HttpClient httpClient,
                      String jwt2IdwsAudience,
                      String kid,
                      CredentialVault idp,
                      CredentialVault signingVault,
                      String scope) {
    this.endpoint = endpoint;
    this.httpClient = httpClient;
    this.jwt2IdwsAudience = jwt2IdwsAudience;
    this.kid = kid;
    this.idp = idp;
    this.signingVault = signingVault;
    this.scope = scope;
  }

  public IdentityToken call(String cpr, String certificateInfo) {

    // if we set cpr every time, the STS does not need to be modified for our needs.
    final String token = makeToken(cpr, CertificateInfo.fromString(certificateInfo));

    final JWTToIdentityTokenRequestDOMBuilder requestBuilder = new OIOSAMLFactory().createJWTToIdentityTokenRequestDOMBuilder();
    requestBuilder.setCPRNumberClaim(cpr); // note something; accept from citizen to know the cpr
    requestBuilder.setJWTToken(token);
    requestBuilder.setSigningVault(signingVault);
    requestBuilder.setAudience(jwt2IdwsAudience);
    final Document requestDocument = requestBuilder.build();
    final String body = XmlUtil.node2String(requestDocument,
                                            false,
                                            false);

    final Response response1 = HttpUtil.issueRequest(httpClient,
                                                     body,
                                                     "",
                                                     endpoint);

    final JWTToIdentityTokenResponse build =
      new OIOSAMLFactory().createJWTToIdentityTokenResponseModelBuilder()
                          .build(XmlUtil.readXml(new Properties(),
                                                 response1.getBody(),
                                                 false));
    if (build.isFault()) {
      throw new RuntimeException("STS error: " + build.getFaultString());
    }

    return build.getIdentityToken();
  }

  private final String issuer = "http://sts-tester";

  private String makeToken(String cpr, CertificateInfo certificateInfo) {
    final String subjectDN = certificateInfo.getSubjectDN().toString();
    final JwtBuilder builder = Jwts.builder();
    builder.signWith(SignatureAlgorithm.RS256, idp.getSystemCredentialPair().getPrivateKey());
    builder.setHeaderParam("kid", kid);
    final Claims claims = Jwts.claims();
    claims.setIssuer(issuer);
    if (subjectDN != null) {
      claims.put("certsubdn", subjectDN);
    }
    claims.put("scope", scope);
    if (cpr != null) {
      claims.put("cpr", cpr);
    }
    claims.setExpiration(new Date(Instant.now().plus(Duration.ofHours(2)).toEpochMilli()));
    builder.setClaims(claims);
    return builder.compact();
  }

}
